#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Viscous-inviscid coupler (quasi-simultaneous coupling)
# Amaury Bilocq

from dart.viscous.Master.DAirfoil import Airfoil
from dart.viscous.Master.DWake import Wake
from matplotlib import pyplot as plt

import fwk
import numpy as np

import tbox.utils as tboxU
import dart.utils as floU
from fwk.coloring import ccolors

class Coupler:
    def __init__(self, _isolver, _vsolver, _boundaryAirfoil, _boundaryWake, _tol, _writer, _bnd):
        self.bnd=_bnd
        self.isolver =_isolver # inviscid solver
        self.vsolver = _vsolver # viscous solver
        
        self.group = [Airfoil(_boundaryAirfoil), Wake(_boundaryWake)] # airfoil and wake python objects
        self.tol = _tol # tolerance of the VII
        self.writer = _writer

        self.tms=fwk.Timers()

    def run(self):
        """Viscous inviscid coupling.
        """

        # Save history.
        cvgMonitoring = []

        # Initialize loop
        it = 0
        converged = 0 # temp
        CdOld = self.vsolver.Cd

        print('+---------------------------- VII Solver begin --------------------------------+')
        print('')

        while converged == 0:

            # Run inviscid solver
            print('+----------------------------- Inviscid solver --------------------------------+')
            self.tms['inv'].start()
            #self.isolver.reset()
            self.isolver.run()  
            self.tms['inv'].stop()

            for n in range(0, len(self.group)):

                for i in range(0, len(self.group[n].boundary.nodes)):
                    
                    # Extract outer flow surface velocities.
                    self.group[n].v[i,0] = self.isolver.U[self.group[n].boundary.nodes[i].row][0]
                    self.group[n].v[i,1] = self.isolver.U[self.group[n].boundary.nodes[i].row][1]
                    self.group[n].v[i,2] = self.isolver.U[self.group[n].boundary.nodes[i].row][2]
                    self.group[n].Me[i] = self.isolver.M[self.group[n].boundary.nodes[i].row]
                    self.group[n].rhoe[i] = self.isolver.rho[self.group[n].boundary.nodes[i].row]
                
            if it==0:
                # Write inviscid pressure distribution on the first iteration
                Cp = floU.extract(self.bnd.groups[0].tag.elems, self.isolver.Cp)
                tboxU.write(Cp, 'Cpinv_airfoil.dat', '%1.5e', ', ', 'x, y, z, Cp', '')

            # Run viscous solver
            print('+------------------------------ Viscous solver --------------------------------+')
            self.tms['visc'].start()
            self.vsolver.run(self.group)
            self.tms['visc'].stop()

            for n in range(0, len(self.group)):
                for i in range(0, self.group[n].nE):
                    self.group[n].boundary.setU(i, self.group[n].u[i])

            error = abs(self.vsolver.Cd - CdOld)/self.vsolver.Cd
            cvgMonitoring.append(error)

            # Check convergence and output coupling iteration.

            print('tms @ it', it, self.tms)
            if it == 84:
              converged = 1

            if error <= self.tol:
              print(self.tms)
              print('')
              print(ccolors.ANSI_GREEN, 'It: {:<3.0f} Cd = {:<6.5f}. Error = {:<2.3f} < {:<2.3f}'
              .format(it, self.vsolver.Cd, np.log10(error), np.log10(self.tol)), ccolors.ANSI_RESET)
              print('')
              """plt.plot(cvgMonitoring[:it+1])
              plt.yscale('log')
              plt.xlabel('Number of iterations (-)')
              plt.ylabel('log[Res] (-)')
              plt.show()"""

              converged = 1
            else:
              print('')
              print(ccolors.ANSI_BLUE, 'It: {:<3.0f} Cd = {:<6.5f}. Error = {:<2.3f} > {:<2.3f}'
              .format(it, self.vsolver.Cd, np.log10(error), np.log10(self.tol)), ccolors.ANSI_RESET)
              print('')

              CdOld = self.vsolver.Cd

            # Prepare next iteration.
            it += 1
            self.vsolver.it += 1

        # Save results

        self.isolver.save(self.writer)    # Inviscid solver files.
        self.vsolver.writeFile()          # Viscous solver files.
        Cp = floU.extract(self.bnd.groups[0].tag.elems, self.isolver.Cp)    # Pressure coefficient.
        tboxU.write(Cp, 'Cp_airfoil.dat', '%1.5e', ', ', 'x, y, z, Cp', '')
        print('\n')
