################################################################################
#                                                                              #
#                            DARTFlo viscous module file                       #
#               CSolver : Boundary layer laws and linear system sparse         #
#                    matrices computation (Residuals, Jacobian)                #
#                                                                              #
# Author: Paul Dechamps                                                        #
# Université de Liège                                                          #
# Date: 2021.11.11                                                             #
#                                                                              #
################################################################################


# ------------------------------------------------------------------------------
#  Imports
# ------------------------------------------------------------------------------
from dart.viscous.Physics.DClosures import Closures

import numpy as np

class flowEquations:

    def __init__(self, _setGradients, _fouScheme) -> None:
        self.setGradients = _setGradients
        self.fouScheme = _fouScheme
        pass
    
    def _blLaws(self, dataCont, iMarker, x) -> np.ndarray:
        """Evaluates the boundary layer laws at one point of the computational domain.

            Args:
            ----
            - dataCont (classType): Data container. Contains the geometry, the solution, BL parameters...

            - iMarker (integer): Id of the point.

            - x (np.ndarray, optional): Prescribed solution, usually a perturbed one for Jacobian computation purposes. Defaults to None.

            Returns:
            -------
            - linSysRes (np.ndarray): Residuals of the linear system at the considered point.
            """

        #print('IC', x)
        
        timeMatrix  = np.empty((dataCont.nVar, dataCont.nVar))
        spaceMatrix = np.empty(dataCont.nVar)
        
        iMarkerPrev, name, xtrF = self.__evalPoint(dataCont, iMarker)
        
        dissipRatio = 0.9 if name == 'wake' else 1 # Wall/wake dissipation length ratio

        dataCont.blPar[iMarker*dataCont.nBlPar+0] = max(x[3] * x[0] * dataCont.Re, 1)  # Reynolds number based on momentum thickness theta

        dataCont.blPar[iMarker * dataCont.nBlPar + 9] = x[0] * x[1]  # deltaStar

        if dataCont.flowRegime[iMarker]==0: 

            # Laminar closure relations.
            dataCont.blPar[iMarker*dataCont.nBlPar+1:iMarker*dataCont.nBlPar+7] = Closures.laminarClosure(x[0], x[1],
                                                                                                         dataCont.blPar[iMarker*dataCont.nBlPar+0],
                                                                                                         dataCont.extPar[iMarker*dataCont.nExtPar+0], name)

        elif dataCont.flowRegime[iMarker]==1:

            # Turbulent closures relations.
            dataCont.blPar[iMarker*dataCont.nBlPar+1:iMarker*dataCont.nBlPar+9] = Closures.turbulentClosure(x[0], x[1],
                                                                                                          x[4], dataCont.blPar[iMarker*dataCont.nBlPar+0],
                                                                                                          dataCont.extPar[iMarker*dataCont.nExtPar+0], name)

        # Rename variables for clarity.
        Ta, Ha, _, uea, Cta = x
        Mea = dataCont.extPar[iMarker*dataCont.nExtPar+0]
        Cda, Cfa, deltaa, Hstara, Hstar2a, Hka, Cteqa, Usa = dataCont.blPar[iMarker * dataCont.nBlPar + 1 : iMarker * dataCont.nBlPar + 9]

        """if iMarker == 1:
          print("Reta",dataCont.blPar[iMarker*dataCont.nBlPar+0])
          print("Hstara",Hstara)
          print("Hstar2a",Hstar2a)
          print("Hka",Hka)
          print("Cfa",Cfa)
          print("Cda",Cda)
          print("deltaa",deltaa)
          print("Usa",Usa)"""
        # Amplification factor for e^n method.
        if name=='airfoil' and xtrF is None and dataCont.flowRegime[iMarker]==0:

            Axa = Closures.amplRoutine(dataCont.blPar[iMarker * dataCont.nBlPar + 6], dataCont.blPar[iMarker * dataCont.nBlPar + 0], Ta)
        
        else:

            Axa=0
            dN_dx=0

        [dTheta_dx, dH_dx, dN_dx, due_dx, dCt_dx], dHstar_dx, dueExt_dx, ddeltaStarExt_dx, c, cExt = self.fouScheme(dataCont, x,
                                                                                                                      iMarkerPrev,
                                                                                                                      iMarker)

        # +------------------------------------------- Equations -----------------------------------------------------------+
        # |																																																									|
        # |																																																									|
        # | Von Karman momentum integral equation 0-th (momentum integral)																									|
        # | --------------------------------------------------------------																									|
        # |																																																									|
        # | H/ue*dTheta/dt + Theta/ue*dH/dt + Theta*H/(ue^2)*due/dt + dTheta/dx + (2+H)*(Theta/ue)*due/dx - Cf/2 = 0				|
        # |																																																									|
        # | 1-st (kinetic-energy integral) moments of momentum equation																											|
        # | -----------------------------------------------------------																											|
        # |																																																									|
        # | (1+H*(1-Hstar))/ue*dTheta/dt + (1-Hstar)*Theta/ue*dH/dt + (2-Hstar*H)*Theta/(ue^2)*due/dt												|
        # | + Theta*(dHstar/dx) + (2*Hstar2 + Hstar*(1-H))*Theta/ue*due_dx - 2*Cd + Hstar*Cf/2 = 0													|
        # |																																																									|
        # | Unsteady e^n method																																															|
        # | -------------------																																															|
        # |																																																									|
        # | dN/dt + dN/dx - Ax = 0 																																													|
        # |																																																									|
        # | Interaction law																																																	|
        # | ---------------																																																	|
        # |																																																									|
        # | due/dt - c*H*dTheta/dt - c*Theta*dH/dt + due/dx-c*(H*dTheta/dx+Theta*dH/dx) + (-due/dx + c*deltaStar)_Old = 0  	|
        # |																																																									|
        # | Unsteady shear lag equation (ignored in the laminar case)																												|
        # | ---------------------------------------------------------																												|
        # |																																																									|
        # | 2*delta/(Us*ue^2)*due/dt + 2*delta/(ue*Ct*Us)*dCt/dt + (2*delta/Ct)*(dCt/dx) 																		|
        # | - 5.6*(Cteq-Ct*dissipRatio)-2*delta*(4/(3*H*Theta)*(Cf/2-((Hk-1)/(6.7*Hk*dissipRatio))^2)-1/ue*due/dx) = 0			|
        # |																																																									|
        # +-----------------------------------------------------------------------------------------------------------------+

        
        # Spacial part
        spaceMatrix[0] = dTheta_dx + (2+Ha-Mea**2) * (Ta/uea) * due_dx - Cfa/2
        spaceMatrix[1] = Ta * dHstar_dx + ( 2*Hstar2a + Hstara * (1-Ha)) * Ta/uea * due_dx - 2*Cda + Hstara * Cfa/2
        if xtrF is not None or dataCont.flowRegime[iMarker] == 1:
            spaceMatrix[2] = 0
        else:
            spaceMatrix[2] = dN_dx - Axa
        spaceMatrix[3] = due_dx -c*(Ha * dTheta_dx + Ta * dH_dx) - dueExt_dx + cExt * ddeltaStarExt_dx
        if dataCont.flowRegime[iMarker] == 1:

            spaceMatrix[4] = (2*deltaa/Cta) * (dCt_dx) - 5.6*(Cteqa - Cta * dissipRatio)
            - 2*deltaa*(4/(3*Ha * Ta) * (Cfa/2 - ((Hka-1)/(6.7*Hka * dissipRatio))**2) - 1/uea * due_dx)

        elif dataCont.flowRegime[iMarker] == 0: spaceMatrix[4] = 0

        # Temporal part
        timeMatrix[0] = [Ha/uea, Ta/uea, 0, Ta * Ha/(uea**2), 0]
        timeMatrix[1] = [(1+Ha * (1-Hstara))/uea, (1-Hstara) * Ta/uea, 0, (2-Hstara * Ha)*Ta/(uea**2), 0]
        timeMatrix[2] = [0, 0, 1, 0, 0]
        timeMatrix[3] = [-c*Ha, -c*Ta, 0, 1, 0]
        if dataCont.flowRegime[iMarker]==1:
            timeMatrix[4] = [0, 0, 0, 2*deltaa/(Usa * uea**2), 2*deltaa/(uea * Cta * Usa)]
        else:
            # Shear lag equation ignored in the laminar portion of the flow
            timeMatrix[4] = [0, 0, 0, 0, 1]
            
        sysRes = np.linalg.solve(timeMatrix, spaceMatrix).flatten()

        """print('dTheta_dx',dTheta_dx)
        print('dH_dx',dH_dx)
        print('due_dx',due_dx)
        print('dCt_dx',dCt_dx)
        print('dueExt_dx',dueExt_dx)
        print('ddstarExt_dx',ddeltaStarExt_dx)
        print('dHstar_dx', dHstar_dx)
        print('2ndTerm', (2+Ha-Mea**2) * (Ta/uea) * due_dx)
        print('third term', Cfa/2)
        print('cteq', Cteqa)

        print('timeMatrix', timeMatrix)
        print('spaceVector', spaceMatrix)
        print('sysRes', sysRes)"""
        
        """sysRes = np.empty(5)
        if dataCont.flowRegime[iMarker] == 0: # Laminar case
          sysRes[0] = (((-((c*Ha*Ta**2)/uea**2) - Ta/uea)*(-2*Cda + (Cfa*Hstara)/2. + Ta*dHstar_dx + 
                      ((2*Hstar2a + (1 - Ha)*Hstara)*Ta*due_dx)/uea))/(-((c*Ha*Ta**2)/uea**3) - Ta/uea**2) + 
                       (((2*c*Ta**2)/uea**2 - (c*Ha*Hstara*Ta**2)/uea**2 + Ta/uea - (Hstara*Ta)/uea)*
                      (-0.5*Cfa + dTheta_dx + ((2 + Ha - Mea**2)*Ta*due_dx)/uea))/(-((c*Ha*Ta**2)/uea**3) - Ta/uea**2) + 
                       (((2*Ta**2)/uea**3 - (Ha*Ta**2)/uea**3)*(cExt*ddeltaStarExt_dx - c*(Ta*dH_dx + Ha*dTheta_dx) + 
                      due_dx - dueExt_dx))/(-((c*Ha*Ta**2)/uea**3) - Ta/uea**2))

          sysRes[1] = ((((c*Ha**2*Ta)/uea**2 + Ha/uea)*(-2*Cda + (Cfa*Hstara)/2. + Ta*dHstar_dx + ((2*Hstar2a + (1 - Ha)*Hstara)*Ta*due_dx)/uea))/
                      (-((c*Ha*Ta**2)/uea**3) - Ta/uea**2) + (((-2*c*Ha*Ta)/uea**2 + (c*Ha**2*Hstara*Ta)/uea**2 - 1/uea - Ha/uea + (Ha*Hstara)/uea)*
                      (-0.5*Cfa + dTheta_dx + ((2 + Ha - Mea**2)*Ta*due_dx)/uea))/(-((c*Ha*Ta**2)/uea**3) - Ta/uea**2) + 
                       ((-((Ha*Ta)/uea**3) + (Ha**2*Ta)/uea**3)*(cExt*ddeltaStarExt_dx - c*(Ta*dH_dx + Ha*dTheta_dx) + 
                      due_dx - dueExt_dx))/(-((c*Ha*Ta**2)/uea**3) - Ta/uea**2))
          
          sysRes[2] = (-Axa + dN_dx)

          sysRes[3] = (-((c*Ta*(-0.5*Cfa + dTheta_dx + ((2 + Ha - Mea**2)*Ta*due_dx)/uea))/((-((c*Ha*Ta**2)/uea**3) - Ta/uea**2)*uea)) - 
                       (Ta*(cExt*ddeltaStarExt_dx - c*(Ta*dH_dx + Ha*dTheta_dx) + due_dx - 
                      dueExt_dx))/((-((c*Ha*Ta**2)/uea**3) - Ta/uea**2)*uea**2))
          
          sysRes[4] = 0

        elif dataCont.flowRegime[iMarker] == 1: # Turbulent case

          sysRes[0] = ((((-2*c*deltaa*Ha*Ta**2)/(Cta*uea**3*Usa) - (2*deltaa*Ta)/(Cta*uea**2*Usa))*
                      (-2*Cda + (Cfa*Hstara)/2. + Ta*dHstar_dx + ((2*Hstar2a + (1 - Ha)*Hstara)*Ta*due_dx)/uea))/
                      ((-2*c*deltaa*Ha*Ta**2)/(Cta*uea**4*Usa) - (2*deltaa*Ta)/(Cta*uea**3*Usa)) + 
                       (((4*c*deltaa*Ta**2)/(Cta*uea**3*Usa) - (2*c*deltaa*Ha*Hstara*Ta**2)/(Cta*uea**3*Usa) + (2*deltaa*Ta)/(Cta*uea**2*Usa) - 
                      (2*deltaa*Hstara*Ta)/(Cta*uea**2*Usa))*(-0.5*Cfa + dTheta_dx + ((2 + Ha - Mea**2)*Ta*due_dx)/uea))/
                      ((-2*c*deltaa*Ha*Ta**2)/(Cta*uea**4*Usa) - (2*deltaa*Ta)/(Cta*uea**3*Usa)) + 
                       (((4*deltaa*Ta**2)/(Cta*uea**4*Usa) - (2*deltaa*Ha*Ta**2)/(Cta*uea**4*Usa))*
                      (cExt*ddeltaStarExt_dx - c*(Ta*dH_dx + Ha*dTheta_dx) + due_dx - dueExt_dx))/
                      ((-2*c*deltaa*Ha*Ta**2)/(Cta*uea**4*Usa) - (2*deltaa*Ta)/(Cta*uea**3*Usa)))
           
          sysRes[1] = ((((2*c*deltaa*Ha**2*Ta)/(Cta*uea**3*Usa) + (2*deltaa*Ha)/(Cta*uea**2*Usa))*
                      (-2*Cda + (Cfa*Hstara)/2. + Ta*dHstar_dx + ((2*Hstar2a + (1 - Ha)*Hstara)*Ta*due_dx)/uea))/
                      ((-2*c*deltaa*Ha*Ta**2)/(Cta*uea**4*Usa) - (2*deltaa*Ta)/(Cta*uea**3*Usa)) - 
                       (2*deltaa*((2*c*Ha*Ta)/uea**2 - (c*Ha**2*Hstara*Ta)/uea**2 + 1/uea + Ha/uea - (Ha*Hstara)/uea)*
                      (-0.5*Cfa + dTheta_dx + ((2 + Ha - Mea**2)*Ta*due_dx)/uea))/
                      (Cta*uea*((-2*c*deltaa*Ha*Ta**2)/(Cta*uea**4*Usa) - (2*deltaa*Ta)/(Cta*uea**3*Usa))*Usa) + 
                       (((-2*deltaa*Ha*Ta)/(Cta*uea**4*Usa) + (2*deltaa*Ha**2*Ta)/(Cta*uea**4*Usa))*
                      (cExt*ddeltaStarExt_dx - c*(Ta*dH_dx + Ha*dTheta_dx) + due_dx - dueExt_dx))/
                      ((-2*c*deltaa*Ha*Ta**2)/(Cta*uea**4*Usa) - (2*deltaa*Ta)/(Cta*uea**3*Usa)))

          sysRes[2] = 0

          sysRes[3] = ((-2*c*deltaa*Ta*(-0.5*Cfa + dTheta_dx + ((2 + Ha - Mea**2)*Ta*due_dx)/uea))/
                      (Cta*uea**2*((-2*c*deltaa*Ha*Ta**2)/(Cta*uea**4*Usa) - (2*deltaa*Ta)/(Cta*uea**3*Usa))*Usa) - 
                      (2*deltaa*Ta*(cExt*ddeltaStarExt_dx - c*(Ta*dH_dx + Ha*dTheta_dx) + due_dx - 
                      dueExt_dx))/(Cta*uea**3*((-2*c*deltaa*Ha*Ta**2)/(Cta*uea**4*Usa) - (2*deltaa*Ta)/(Cta*uea**3*Usa))*Usa))


          sysRes[4] = ((2*c*deltaa*Ta*(-0.5*Cfa + dTheta_dx + ((2 + Ha - Mea**2)*Ta*due_dx)/uea))/
                      (uea**3*((-2*c*deltaa*Ha*Ta**2)/(Cta*uea**4*Usa) - (2*deltaa*Ta)/(Cta*uea**3*Usa))*Usa) + 
                       ((-((c*Ha*Ta**2)/uea**3) - Ta/uea**2)*(-5.6*(Cteqa - Cta*dissipRatio) + (2*deltaa*dCt_dx)/Cta - 
                      2*deltaa*((4*(Cfa/2. - (0.02227667631989307*(-1 + Hka)**2)/(dissipRatio**2*Hka**2)))/(3.*Ha*Ta) - due_dx/uea)))/
                      ((-2*c*deltaa*Ha*Ta**2)/(Cta*uea**4*Usa) - (2*deltaa*Ta)/(Cta*uea**3*Usa)) + 
                       (2*deltaa*Ta*(cExt*ddeltaStarExt_dx - c*(Ta*dH_dx + Ha*dTheta_dx) + due_dx - 
                      dueExt_dx))/(uea**4*((-2*c*deltaa*Ha*Ta**2)/(Cta*uea**4*Usa) - (2*deltaa*Ta)/(Cta*uea**3*Usa))*Usa))"""


        return sysRes


    def __evalPoint(self, dataCont, iMarker) -> tuple:
        """Evaluates where the point is, the adjactent points
            and the state of transition on the corresponding side

            Args:
            ----
            - iMarker (integer): Id of the cell to evaluate.

            - bNodes (np.ndarray): Boundary nodes of the computational domain.

            - xtrIn (np.ndarray): Current transiton location on [upper, lower] sides.

            - xtrFIn (np.ndarray): Forced transition state on [upper, lower] sides.

            - nMarker ([type]): Number of cells in the computational domain.

            Returns:
            -------
            - tuple: Id of the adjacent points, name of the corresponding group ('airfoil', 'wake'),
                transition state on the corresponding side.
            """
        iMarkerPrev = 0 if iMarker == dataCont.bNodes[1] else iMarker - 1

        if iMarker < dataCont.bNodes[1]:
            name='airfoil'
            xtrF= dataCont.xtrF[0]

        elif iMarker < dataCont.bNodes[2]:
            name = 'airfoil'
            xtrF = dataCont.xtrF[1]

        else:
            name = 'wake'
            xtrF = None
        return iMarkerPrev, name, xtrF


class sysMatrix(flowEquations):
    def __init__(self, _nMarker, _nMarkers, setGradients, fouScheme) -> None:
        flowEquations.__init__(self, setGradients, fouScheme)
        self._nMarker = _nMarker
        self._nMarkers = _nMarkers
        pass

    def buildJacobian(self, dataCont, marker, sysRes, order = 1, eps = 1e-10) -> np.ndarray:
        """Contruct the Jacobian used for solving the liner system.
        
            Args:
            ----
            - dataCont (classType): Data container (solution, BL parameters, invicid flow parameters, geometry...)

            - marker (integer, optional): Cell id on the computational domain. Default : -1 for whole domain.

            - order (unsigned int, optional): Order of the discretization used to obtain the Jacobian (1st or 2nd order implemented).
                                        If order == 1, linSysRes has to be specifed

            - linSysRes (np.ndarray, optional): Residuals of the linear system at the current iteration. Used (and mandatory) if order == 1.

            - eps (double, optional): Perturbation on the variables. Default : 10^(-8).

            Returns
            -------

            - JacMatrix (np.ndarray): Jacobian of the system. 

            """
        
        # Build Jacobian for one point 
        if order == 1 and sysRes is not None:
            
            JacMatrix = np.zeros((dataCont.nVar, dataCont.nVar))
            xUp = dataCont.u[marker*dataCont.nVar : (marker+1)*dataCont.nVar].copy()

            for iVar in range(dataCont.nVar):
                varSave = xUp[iVar]
                xUp[iVar] += eps

                JacMatrix[:,iVar] = (self._blLaws(dataCont, marker, x=xUp) - sysRes) / eps

                xUp[iVar] = varSave
            return JacMatrix

        elif order == 2:
            JacMatrix = np.zeros((dataCont.nVar, dataCont.nVar))
            xUp = dataCont.u[marker*dataCont.nVar : (marker+1)*dataCont.nVar].copy()
            xDwn = dataCont.u[marker*dataCont.nVar : (marker+1)*dataCont.nVar].copy()

            for iVar in range(dataCont.nVar):
                varSave = xUp[iVar]
                xUp[iVar] += eps
                xDwn[iVar] -= eps

                JacMatrix[:,iVar] = (self._blLaws(dataCont, marker, x = xUp) - self._blLaws(dataCont, marker, x = xDwn)) / (2*eps)

                xUp[iVar] = varSave
                xDwn[iVar] = varSave
            return JacMatrix

        else:
            raise RuntimeError('Only first and second orders Jacobian can be computed.')


    def buildResiduals(self, dataCont, marker) -> np.ndarray:
        """Construct residuals of the linear system.
        
            Args:
            ----
            
            - dataCont (classType): Data container (solution, BL parameters, invicid flow parameters, geometry...)

            - marker (int, optional): Cell id on the computational domain. default: -1 for whole domain.
            """

        # Build Residual for one point
        return self._blLaws(dataCont, marker, x=dataCont.u[marker*dataCont.nVar:(marker+1)*dataCont.nVar])